/**************************************************************************************************
MiniSat -- Copyright (c) 2003-2005, Niklas Een, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/
#include <iostream>


#include "Constraints.h"
#include "Solver.h"
#include "Sort.h"

using namespace std;

//=================================================================================================
// Helpers:


void removeWatch(vec<Constr*>& ws, Constr* elem)
{
    int j = 0;
    for (; ws[j] != elem  ; j++) assert(j < ws.size());
    for (; j < ws.size()-1; j++) ws[j] = ws[j+1];
    ws.pop();
}
void printLits( const vec<Lit>& ps_)
{
  cout<<"Lits:   ";
    for (int i = 0; i < ps_.size(); i++){
      cout<<index(ps_[i])<<' ';
	}
	cout<<endl;
}


//=================================================================================================
// Clause_ms constraint:


// Returns FALSE if top-level conflict detected (must be handled); TRUE otherwise.
// 'out_clause' may be set to NULL if clause is already satisfied by the top-level assignment.
//


//bool Clause_ms_new(Solver& S, const vec<Lit>& ps_, bool learnt, Clause_ms*& out_clause)

bool Clause_ms_new(Solver& S, const vec<Lit>& ps_, bool learnt, Clause_ms*& out_clause)
{
    vec<Lit>    qs;
    if (&out_clause != NULL) out_clause = NULL;

    if (!learnt){
        assert(S.decisionLevel() == 0);
        ps_.copyTo(qs);             // Make a copy of the input vector.

        // Remove false literals:
        for (int i = 0; i < qs.size();){
            if (S.value(qs[i]) != l_Undef){
                if (S.value(qs[i]) == l_True)
                    return true;    // Clause always true -- don't add anything.
                else
                    qs[i] = qs.last(),
                    qs.pop();
            }else
                i++;
        }

        // Remove duplicates:
        sortUnique(qs);
        for (int i = 0; i < qs.size()-1; i++){
            if (qs[i] == ~qs[i+1])
                return true;        // Clause always true -- don't add anything.
        }
    }
    const vec<Lit>& ps = learnt ? ps_ : qs;     // 'ps' is now the (possibly) reduced vector of literals.

    if (ps.size() == 0)
        return false;
    else if (ps.size() == 1)
        return S.enqueue(ps[0]);
    else{
        // Allocate clause:
        assert(sizeof(Lit)   == sizeof(unsigned));
        assert(sizeof(float) == sizeof(unsigned));
        void*   mem = xmalloc<char>(sizeof(Clause_ms) + sizeof(unsigned)*(ps.size() + (int)learnt));
        Clause_ms* c   = new (mem) Clause_ms;

        c->size_learnt = (int)learnt | (ps.size() << 1);
        for (int i = 0; i < ps.size(); i++) c->data[i] = ps[i];
        if (learnt) c->activity() = 0.0;

        // For learnt clauses only:
        if (learnt){
            // Put the second watch on the literal with highest decision level:
            int     max_i = 1;
            int     max   = S.level[var(ps[1])];
            for (int i = 2; i < ps.size(); i++)
                if (S.level[var(ps[i])] > max)
                    max   = S.level[var(ps[i])],
                    max_i = i;
            c->data[1]     = ps[max_i];
            c->data[max_i] = ps[1];

            // Bumping:
            S.claBumpActivity(c); // (newly learnt clauses should be considered active)

            S.stats.learnts++;
            S.stats.learnts_literals += c->size();
        }else{
            S.stats.clauses++;
            S.stats.clauses_literals += c->size();
        }

        // Store clause:
        S.watches[index(~c->data[0])].push(c);
        S.watches[index(~c->data[1])].push(c);
        if (&out_clause != NULL) out_clause = c;
		/*		if(learnt)
		  cout<<"Learned ";
		c->print();
		*/
        return true;
    }
}
/*

bool Clause_ms_new(Solver& S, const vec<Lit>& ps_, bool learnt, Clause_ms*& out_clause)
{

    vec<Lit>    qs;
    if (&out_clause != NULL) out_clause = NULL;

    if (!learnt){
        assert(S.decisionLevel() == 0);
        ps_.copyTo(qs);             // Make a copy of the input vector.

        // Remove false literals:
        for (int i = 0; i < qs.size();){
            if (S.value(qs[i]) != l_Undef){
                if (S.value(qs[i]) == l_True)
                    return true;    // Clause_ms always true -- don't add anything.
                else
                    qs[i] = qs.last(),
                    qs.pop();
            }else
                i++;
        }

        // Remove duplicates:
        sortUnique(qs);
        for (int i = 0; i < qs.size()-1; i++){
            if (qs[i] == ~qs[i+1])
                return true;        // Clause_ms always true -- don't add anything.
        }
    }
    const vec<Lit>& ps = learnt ? ps_ : qs;     // 'ps' is now the (possibly) reduced vector of literals.

    if (ps.size() == 0)
        return false;
    else if (ps.size() == 1)
        return S.enqueue(ps[0]);
    else{
        // Allocate clause:
        assert(sizeof(Lit)   == sizeof(unsigned));
        assert(sizeof(float) == sizeof(unsigned));
        void*   mem = xmalloc<char>(sizeof(Clause_ms) + sizeof(unsigned)*(ps.size() + (int)learnt));
        Clause_ms* c   = new (mem) Clause_ms;

        c->size_learnt = (int)learnt | (ps.size() << 1);
        for (int i = 0; i < ps.size(); i++) c->data[i] = ps[i];
        if (learnt) c->activity() = 0.0;

        // For learnt clauses only:
        if (learnt){
            // Put the second watch on the literal with highest decision level:
            int     max_i = 1;
            int     max   = S.level[var(ps[1])];
            for (int i = 2; i < ps.size(); i++)
                if (S.level[var(ps[i])] > max)
                    max   = S.level[var(ps[i])],
                    max_i = i;
            c->data[1]     = ps[max_i];
            c->data[max_i] = ps[1];

            // Bumping:
            S.claBumpActivity(c); // (newly learnt clauses should be considered active)

            S.stats.learnts++;
            S.stats.learnts_literals += c->size();
        }else{
            S.stats.clauses++;
            S.stats.clauses_literals += c->size();
        }

        // Store clause:
        S.watches[index(~c->data[0])].push(c);
        S.watches[index(~c->data[1])].push(c);
        if (&out_clause != NULL) out_clause = c;
				if(learnt)
		cout<<"Learned ";
				c->print();
								
        return true;
    }
}
*/

bool Clause_ms::locked(const Solver& S) const {
    return (const Clause_ms *)S.reason[var(data[0])] == this; }


void Clause_ms::remove(Solver& S, bool just_dealloc)
{
    if (!just_dealloc){
        removeWatch(S.watches[index(~data[0])], this);
        removeWatch(S.watches[index(~data[1])], this); }

    if (learnt()){
        S.stats.learnts--;
        S.stats.learnts_literals -= size();
    }else{
        S.stats.clauses--;
        S.stats.clauses_literals -= size();
    }

    xfree(this);
}


// Can assume everything has been propagated! (esp. the first two literals are != l_False, unless
// the clause is binary and satisfied, in which case the first literal is true)
// Returns True if clause is satisfied (will be removed), False otherwise.
//
bool Clause_ms::simplify(Solver& S)
{
    assert(S.decisionLevel() == 0);
    for (int i = 0; i < size(); i++){
        if (S.value(data[i]) == l_True)
            return true;
    }
    return false;
}


// 'p' is the literal that became TRUE
bool Clause_ms::propagate(Solver& S, Lit p, bool& keep_watch)
{
    // Make sure the false literal is data[1]:
    Lit     false_lit = ~p;
    if (data[0] == false_lit)
        data[0] = data[1], data[1] = false_lit;
    assert(data[1] == false_lit);

    // If 0th watch is true, then clause is already satisfied.
    if (S.value(data[0]) == l_True){
        keep_watch = true;
        return true; }

    // Look for new watch:
    for (int i = 2; i < size(); i++){
        if (S.value(data[i]) != l_False){
            data[1] = data[i], data[i] = false_lit;
            S.watches[index(~data[1])].push(this);
            return true; } }

    // Clause_ms is unit under assignment:
    keep_watch = true;
	//    cout<<"Clause propagate ";
	//    print();
	//	cout<<"  enque "<<index(data[0])<<endl;
    return S.enqueue(data[0], this);
}


// Can assume 'out_reason' to be empty.
// Calculate reason for 'p'. If 'p == lit_Undef', calculate reason for conflict.
//
void Clause_ms::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
    assert(p == lit_Undef || p == data[0]);
    for (int i = ((p == lit_Undef) ? 0 : 1); i < size(); i++)
        assert(S.value(data[i]) == l_False),
        out_reason.push(~data[i]);
    if (learnt()) S.claBumpActivity(this);
}


//=================================================================================================
// AtMost constraint -- An example of extending MiniSat:


// Pre-condition: All variables must be distinct and unbound. (To lift this pre-condition,
// any fact immediately derivable from the new constraint should be enqueued by 'enqueue()'.
// If the constraint is already satisfied under the current top-level assignment, it should
// not be constructed at all.)
//
bool AtMost_new(Solver& S, const vec<Lit>& ps, int n, AtMost*& out)
{
    assert(S.decisionLevel() == 0);

    void* mem    = (void*)xmalloc<char>(sizeof(AtMost) + sizeof(Lit) * ps.size());
    out          = new (mem) AtMost;
    out->size    = ps.size();
    out->n       = n;
    out->counter = 0;
    for (int i = 0; i < ps.size(); i++) out->lits[i] = ps[i];
    for (int i = 0; i < ps.size(); i++) S.watches[index(ps[i])].push(out);

    return true;
}


void AtMost::remove(Solver& S, bool just_dealloc) {
    if (!just_dealloc)
        for (int i = 0; i < size; i++)
            removeWatch(S.watches[index(lits[i])], this);
    xfree(this);
}


bool AtMost::simplify(Solver& S) {
    return false; }


bool AtMost::propagate(Solver& S, Lit p, bool& keep_watch)
{
    keep_watch = true;
    if (counter == n) return false;

    counter++;
    S.undos[var(p)].push(this);

    // If no more can be true, enqueue the negation of the rest:
    if (counter == n)
        for (int i = 0; i < size; i++)
            if (!S.enqueue(~lits[i], this))
                return false;
    return true;
}


void AtMost::undo(Solver& S, Lit p) {
    counter--; }


void AtMost::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
    int     c = (p == lit_Undef) ? -1 : 0;
    for (int i = 0; i < size; i++){
        if (S.value(lits[i]) == l_True){
            out_reason.push(lits[i]);
            if (++c == n) return;
        }
    }
}

//=================================================================================================
// Rule_ms constraint:


// Returns FALSE if top-level conflict detected (must be handled); TRUE otherwise.
// 'out' may be set to NULL if clause is already satisfied by the top-level assignment.
//


bool Rule_ms_new  (Solver& S, const vec<Lit>& ps_, Rule_ms*& out, bool chrule)
{
  int s=ps_.size();
    if (&out != NULL) out = NULL;
	assert(S.decisionLevel() == 0);

	if (S.value(ps_[0]) == l_False)
		return true; //since head atom is already known to be False
	  //ARC is not needed; UP will take care of everything
	 
	Lit head=ps_[0];
	//	assert(sign(head));
	bool headSelfDep=false;
	// Remove false literals:
    vec<Lit>    qs;
	ps_.copyTo(qs);             // Make a copy of the input vector.
	// Remove false literals:
	for (int i = 1; i < qs.size();){
	  if (S.value(qs[i]) != l_Undef){
		if (S.value(qs[i]) == l_True )
		  return true;    // Body always false hence we do not add this rule to headof
		else
		  qs[i] = qs.last(),
			qs.pop();
	  }else{
		if(head==qs[i])// case a:-...not a...
		  //we need to remember that there is a self dependency
		  headSelfDep=true;

		i++;
	  }
	}
	// Remove duplicates:
	sortUnique(qs);
	for (int i = 0; i < qs.size()-1; i++){
	  if (qs[i] == ~qs[i+1])
		return true;        // Clause_ms always true -- don't add anything.
	}
	if(headSelfDep){
	  //add head as a first element [0] and push the first element in the end of the list
	  qs.push(qs[0]);
	  qs[0]=head;
	}//otherwise we need to find head and replace it with first element by placing head first

	for (int i=0;i<qs.size();i++){
	  if (qs[i]==head){
	    qs[i]=qs[0];
		qs[0]=head;
	  }
	}

    const vec<Lit>& ps =  qs;     // 'ps' is now the (possibly) reduced vector of literals.
	//
	//now we increase headcount
	//
	S.hcountInc(head);


    if (ps.size() == 0){ //i do not think we can get to this case
        return false;
	}
    else if (ps.size() == 1){
	  if(!chrule){
		//		assert(s==1);
		//		if(s!=1)
		// printLits(S);
		//		cout<<s<<" enque head";
		//	  exit(0);
	  
		return S.enqueue(ps[0]);//if body is true then head atom is propagates
	  }
	  else {
		//		cout<<"add count";
		//	  exit(0);

		//		S.hcountInc(ps[0]);	  
	  }
	}
    else{
        // Allocate clause:
        assert(sizeof(Lit)   == sizeof(unsigned));
        assert(sizeof(float) == sizeof(unsigned));
        void*   mem = xmalloc<char>(sizeof(Rule_ms) + sizeof(unsigned)*(ps.size()));
        Rule_ms* c   = new (mem) Rule_ms;


        for (int i = 0; i < ps.size(); i++){
		  c->lits[i] = ps[i];
		}
		c->setActive(true);

		c->setSize(ps.size());


		S.stats.rules++;

		//																c->print();
		S.addRuleToHead(ps[0],c);
		for (int i = 1; i < ps.size(); i++)
		  // Store rule constraints:
		  //here we go through all body lits and add constraint
		  S.watches[index(ps[i])].push(c);

        if (&out != NULL) out = c;


	}
	return true;
}




void Rule_ms::remove(Solver& S, bool just_dealloc)
{
    if (!just_dealloc){
	  for (int i = 1; i < size; i++){
		  // Store rule constraints:
		  //here we go through all body lits and add constraint
		removeWatch(S.watches[index(lits[i])],this);
	  }
	  removeWatch(S.hrules[var(lits[0])],this);
	}

	S.stats.rules--;
    xfree(this);
}


// Can assume everything has been propagated! (esp. the first two literals are != l_False, unless
// the clause is binary and satisfied, in which case the first literal is true)
// Returns True if clause is satisfied (will be removed), False otherwise.
//
bool Rule_ms::simplify(Solver& S)
{
    return false;
}


// 'p' is the literal that became TRUE
bool Rule_ms::propagate(Solver& S, Lit p, bool& keep_watch)
{
  keep_watch=true;
  if(!active)
	return true;
  else
	active=false;
  S.hcountDec(lits[0]);
	//	here we need to store p as a reason for this headCount going down
  S.addLitToReason(lits[0],p);
  S.undos[var(p)].push(this);
    
  if(S.hcountVal(lits[0])==1&& S.value(lits[0]) == l_True){//backchain true is applicable
	//we go through the list of all rules and find one that is active
	//only one rule is active for the head and head value is true
	vec<Constr*> rules;
	S.headRules (lits[0], rules);
	Rule_ms* r;
	for(int i=0; i<rules.size();i++){
	  r = (Rule_ms*)rules[i];
	  if(r->active){//this rule is active then we can enque all its body
		for(int j=1; j<r->size;j++){
		  propagated.push(~r->lits[j]);
		  if(!S.enqueue(~r->lits[j], this))
			 return false;
		}
		return true;
	  }
	}
	//	assert(false);//we should never reach here

  }
  
  else
  if(S.hcountVal(lits[0])==0){//enque compelement of head 
	//    cout<<"Rule propagate "<<endl;
	//    print();
	//	cout<<"  enque "<<index(~lits[0])<<endl;
	//	setConflictLiteral(lits[0]);
	//propagated.push(~lits[0]);
    return S.enqueue(~lits[0], this);
	
  }


  return true;
}

//
//undo for Rule_ms, makes the rule active again:
// rule is 
// 1) active again; 
// 2) headliteralIncremented and 
// 3) removed from reasons for the head literal
//
void Rule_ms::undo(Solver& S, Lit p) {
  //  conf=-1;
  active=true;
  //    cout<<"undo  ";
  //  print();
  S.hcountInc(lits[0]);
  //  cout<<" hcount "<<S.hcountVal(lits[0]);
  S.popFromReason(lits[0]);// here i suppose that this lit is last in reaosn
  propagated.clear();
  //this seem to be the case, but may be we will need to modify this and look through the whole reason list
}


// Can assume 'out_reason' to be empty.
// Calculate reason for 'p'. If 'p == lit_Undef', calculate reason for conflict.
//
void Rule_ms::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
 
  //  assert(p == lit_Undef || p == ~lits[0]);
  // three cases p==lit_undef
  //p==~lits[0]
  //p==~ of some literal in the body
  vec<Lit> r;
  S.headReason(lits[0],r);
  for (int i = 0; i < r.size(); i++){
	assert(S.value(r[i]) == l_True),
	  out_reason.push(r[i]);
  }
  //if(p == lit_Undef)
  //out_reason.push(lits[0]);//if 
  
  if(propagated.size()>0||p == lit_Undef)
	out_reason.push(lits[0]);//if 
  if(propagated.size()>0&&p == lit_Undef){
	//	  cout<<"We are in claculating conflict reason"<<endl;
	for(int i=0;i<propagated.size();i++)
	  if(S.value(propagated[i])!=l_True){
		//	  cout<<"Found atom"<<endl;
		out_reason.push(~propagated[i]);
		break;
	  }
	
  }
	
}

void Rule_ms::print(Solver& S)
{
  cout<<"r:   ";
  if(active) 
	cout<< " active ";
  else 
	cout<< " not active ";
  for (int i = 0; i < size; i++){
	cout<<index(lits[i])<<' ';
	if(S.value(lits[i])==l_True)
		cout<<":T ";
	else if(S.value(lits[i])==l_False)
		cout<<":F ";
	else
	  cout<<":U ";

  }
  cout<<endl;
}
void Rule_ms::print()
{
  cout<<"r:   ";
  if(active) 
	cout<< " active ";
  else 
	cout<< " not active ";
  for (int i = 0; i < size; i++){
	cout<<index(lits[i])<<' ';
	//		cout<<S.value(lits[i])<<' ';
  }
  cout<<endl;
}
void Clause_ms::print()
{  cout<<"cl:  ";
    for (int i = 0; i < size(); i++){
      cout<<index(data[i])<<' ';
	}
	cout<<endl;
}

