/**************************************************************************************************
MiniSat -- Copyright (c) 2003-2005, Niklas Een, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#ifndef Constraints_h
#define Constraints_h

#include "SolverTypes.h"


//=================================================================================================
// Constraint abstraction:


class Solver;

struct Constr {
    virtual void remove    (Solver& S, bool just_dealloc = false) = 0;
    virtual bool propagate (Solver& S, Lit p, bool& keep_watch) = 0;    // ('keep_watch' is set to FALSE beftore call to this method)
    virtual bool simplify  (Solver& S) { return false; };
    virtual void undo      (Solver& S, Lit p) { };
    virtual void calcReason(Solver& S, Lit p, vec<Lit>& out_reason) = 0;
    virtual void print() = 0;

    virtual ~Constr(void) { };  // (not used, just to keep the compiler happy)
};


//=================================================================================================
// Clause_mss:


class Clause_ms : public Constr {
    unsigned    size_learnt;
    Lit         data[0];

public:
    int  size        (void)      const { return size_learnt >> 1; }
    bool learnt      (void)      const { return size_learnt & 1; }
    Lit  operator [] (int index) const { return data[index]; }

    // Constructor -- creates a new clause and add it to watcher lists. 
    friend bool Clause_ms_new(Solver& S, const vec<Lit>& ps, bool learnt, Clause_ms*& out_clause);

    // Learnt clauses only:
    bool    locked  (const Solver& S) const;
    float&  activity(void) const { return *((float*)&data[size()]); }

    // Constraint interface:
    void remove    (Solver& S, bool just_dealloc = false);
    bool propagate (Solver& S, Lit p, bool& keep_watch);
    bool simplify  (Solver& S);
    void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
	void print();
};


//=================================================================================================
// AtMost:


class AtMost : public Constr {
    int     n;
    int     counter;
    int     size;
    Lit     lits[0];

public:
    // Constructor -- creates a new AtMost-constraint and add it to watcher lists.
    friend bool AtMost_new(Solver& S, const vec<Lit>& ps, int n, AtMost*& out);

    // Constraint interface:
    void remove    (Solver& S, bool just_dealloc = false);
    bool propagate (Solver& S, Lit p, bool& keep_watch);
    bool simplify  (Solver& S);
    void undo      (Solver& S, Lit p);
    void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
	void print(){};
};

//=================================================================================================
// ARC: Stopped here.


class Rule_ms : public Constr {
  bool    active;
  int     size;
  // Lit     conf; //conflict literal 
  vec<Lit> propagated;     
  Lit     lits[0];//ALWAYS MAKE SURE THAT THIS IS THE LAST FIELD!

public:
    // Constructor -- creates a new Rule_ms-constraint and adds it to watcher lists.
    friend bool Rule_ms_new(Solver& S, const vec<Lit>& ps,  Rule_ms *& out, bool sth);
    int  getSize        (void)  { return size; }
    void  setSize        (int s)  { size=s; }
    Lit  getLit        (int s)  { assert(s>-1&&s<size);
	                             return lits[s]; }
    //void  setConflictLiteral        (Lit s)  { conf=s; }
    //Lit  getConflictLiteral ()  { return conf; }
    void  setActive        (bool b)  { active=b; }
	bool  getActive        ()  { return active; }
    // Constraint interface:
    void remove    (Solver& S, bool just_dealloc = false);
    bool propagate (Solver& S, Lit p, bool& keep_watch);
    bool simplify  (Solver& S);
    void undo      (Solver& S, Lit p);
    void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
	void print();
	void print(Solver& S);
};


//=================================================================================================
#endif
